Package: stringmagic
Type: Package
Title: Character String Operations and Interpolation, Magic Edition
Version: 1.3.0
Authors@R: 
  c(person(given = "Laurent R", family = "Berge", role = c("aut", "cre"), email = "laurent.berge@u-bordeaux.fr",
         comment = c(ORCID = "0000-0003-2736-1917")),
    person(given = "Morgan", family = "Jacob", role = "ctb"))
Imports: Rcpp(>= 1.0.5), utils, stats
Suggests: knitr, rmarkdown, data.table
LinkingTo: Rcpp
Description: Performs complex string operations compactly and efficiently. Supports string interpolation jointly with over 50 string operations. Also enhances regular string functions (like grep() and co). See an introduction at <https://lrberge.github.io/stringmagic/>.
License: GPL (>= 2)
Encoding: UTF-8
BugReports: https://github.com/lrberge/stringmagic/issues
URL: https://lrberge.github.io/stringmagic/,
        https://github.com/lrberge/stringmagic
VignetteBuilder: knitr
RoxygenNote: 7.3.2
NeedsCompilation: yes
Packaged: 2026-08-03 13:42:56 UTC; berge028
Author: Laurent R Berge [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2736-1917>),
  Morgan Jacob [ctb]
Maintainer: Laurent R Berge <laurent.berge@u-bordeaux.fr>
Repository: CRAN
Date/Publication: 2026-08-04 10:30:09 UTC
